#!/bin/bash -f
#---------------------------------------------------------------------
# Script to display unicode characters representing the level of
# some indicator.
#
# Designed to work with the wonderful byobu(1) but can be run
# stand-alone.
#---------------------------------------------------------------------
#
# Copyright (C) 2018 F-Hauri.ch
# This work is a total rewrite based on 
#
# byobu-ulevel ``Copyright (C) 2011 Canonical Ltd.
#
#                Author: James Hunt <james.hunt@canonical.com>''
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------
# Themes with 2 values.
#
# Two-value themes are handled differently to other n-value types of
# theme: the first array entry in each theme is generally some unfilled
# glyph, denoting an "off" value and the second value is the filled
# version of the unfilled glyph and denotes an "on" value.  Note that
# you can always change the ordering of these values using the 'invert'
# command-line option.

default_decimal_places=2
a11y=false
a11y_variable=BYOBU_A11Y
debug_enabled=false
newline=
list=false
quiet=false
invert=false
reverse=false
script_name=${0##*/}
min_default=0
max_default=100
width_default=5
zero_as_space=false
theme_default=vbars_8
permissive=false

declare -A themes="(
    [circles_2]='○ ●' [diamonds_2]='◇ ◆'      [flags_2]='⚐ ⚑'  [hearts_2]='♡ ♥' 
    [squares_2]='◻ ◼' [squares_small_2]='◽ ◾' [stars_2]='☆ ★'  [faces_2]='☹ ☺' 

    [vdots_thick_4]='⣀ ⣤ ⣶ ⣿' [vdots_thin_4]='⢀ ⢠ ⢰ ⢸'
    [fractions_4]='¼ ½ ¾ ¹'      [quadrants_4]='◔ ◑ ◕ ●'   [shades_4]='░ ▒ ▓ █' 

    [circles_5]='◦ ○ ◎ ◉ ●'                              [dice_6]='⚀ ⚁ ⚂ ⚃ ⚄ ⚅' 

    [hbars_8]='▏ ▎ ▍ ▌ ▋ ▊ ▉ █'  [vbars_8]='▁ ▂ ▃ ▄ ▅ ▆ ▇ █' 
    [circle_number_10]='➀ ➁ ➂ ➃ ➄ ➅ ➆ ➇ ➈ ➉'
    [solid_numbers_a_10]='➊ ➋ ➌ ➍ ➎ ➏ ➐ ➑ ➒ ➓' 
    [solid_numbers_b_10]='❶ ❷ ❸ ❹ ❺ ❻ ❼ ❽ ❾ ❿' 
)"

debug() { $debug_enabled && echo "DEBUG: $*" ;}
error() { echo >&2 "$0 ERROR: $*" ;}
die()   { error "$*" ;exit 1 ;}
check_a11y() { local result=${!a11y_variable};[ "$result" ] && a11y=y ;}
usage() { cat <<-EOT
	Description: Display unicode characters representing the relative
	             level of some indicator value within a range.

	Usage: $script_name [options] -c <current_num>
	       $script_name [options] <current_num>
	       $script_name <current_num>

	Options:

	 -a         : Accessibility mode: only output ASCII.
	              (Also enabled if variable '$a11y_variable' set).
	 -b         : Display current value as space if zero, rather than lowest
	              'value' of theme.
	 -c <num>   : Current value of your indicator.
	 -d         : Enable debug output.
	 -e <int>   : Number of decimal places to use for accessibility mode
	              (default=$default_decimal_places).
	 -h         : Show this help.
	 -i         : Invert colour scheme (rating themes only).
	 -l         : List available themes. If '-t' also specified,
	              show all values for specified theme.
	 -m <num>   : Minimum value (default=$min_default).
	 -n         : Supress output of newline character.
	 -p         : Permissive mode - if current value out of bounds, set it
	              to the nearest bound (min or max).
	 -q         : Suppress messages (requires '-t').
	 -r         : Reverse 'direction' of display (rating theme only).
	 -t <theme> : Name of theme (default=$theme_default).
	 -u <chars> : Specify a user theme (2 or more values).
	 -w <int>   : Width of rating theme (default=$width_default).
	 -x <num>   : Maximum value (default=$max_default).

	Examples:

	  # Display character representing 27% using default theme.
	  $script_name 27

	  # As above.
	  $script_name -c 27

	  # Example showing floating-point and negative values.
	  $script_name -c 1.100001 -m -5.00234 -x 2.71828 -t dice_6

	  # Use accessibility mode to display a percentage value
	  # (rounded to nearest percentage)
	  $script_name -m -22.613 -x 5.00212 -c 0.10203 -a -e 0

	  # Display value using a "rating theme" (displayed left-to-right).
	  $script_name -c 83 -t stars_2

	  # Display right-to-left inverted "rating theme".
	  $script_name -c 60 -t diamonds_2 -ri

	  # Display all glyphs in 'solid_numbers_a_10' theme.
	  $script_name -l -t solid_numbers_a_10

	  # Display a user-specified rating theme 10 glyphs wide.
	  $script_name -c 666.321 -m -273.15 -x 1370 -u "· ☢" -w 10

	  # A multi-element user theme (this prints 'e').
	  $script_name -c 50 -u "a b c d e f g h i j"

	Notes:

	  - Arguments of type "<int>" denote an integer value, whereas arguments
	    of type "<num>" denotes either an integer or a floating-point
	    number.
	  - The final '_<number>' in a theme name denotes the number of glyphs
	    in it.
	  - "Rating themes" are those with only 2 values.
	  - The <chars> argument to '-u' must contain space-delimited
	    characters.
	EOT
}

# prepare bc for long-running subprocess
# find next free fd
nfd() {
    local var=$1 nfd=3;
    while [ -e /dev/fd/$nfd ] ;do ((nfd++)) ;done
    printf -v $var %d $nfd
}
# run bc in background and create `myBc` to deal with.
runBackBc() {
    local bctmpdir
    read bctmpdir < <(mktemp -d /dev/shm/byobu-ulevel-bc-XXXXXXX)
    nfd BCIN
    mkfifo $bctmpdir/bcout
    eval "exec $BCIN> >(LANG=C exec stdbuf -o0 bc -l >$bctmpdir/bcout 2>&1)"
    nfd BCOUT
    eval "exec $BCOUT<$bctmpdir/bcout"
    rm $bctmpdir/bcout
    rmdir $bctmpdir
    eval "myBc() { echo >&$BCIN \"\$1\";read -u $BCOUT \$2 ;}"
}
bc_test() {
    local expr="$*" ans
    myBc "if ( $expr ) 0 else 1" ans
    return $ans
}
assert() {
    local expr="$1" str="$2"
    debug "assert: expr='$expr'"
    bc_test "$expr" && return
    die "$str"
}
list_theme() {
    printf "\n%s\n\n" "${themes[$1]}"
}
list_themes() {
    printf "%s\n" ${!themes[@]}
}

theme_valid() {
    local _thm
    for _thm in ${!themes[@]};do
        [ "$_thm" = "$1" ] && return 0
    done
    return 1
}
show_rating_theme() {
    # local -a theme="($1)"
    local -a theme="(${themes[$1]})"
    local min="$2" max="$3" current="$4" percent="$5"
    local on off
    if $invert ;then
	on=0
	off=1
    else
	on=1
	off=0
    fi
    
    debug "width=$width"
    debug "percent=$percent"
    
    myBc "scale=4;100/${width};scale=20" percent_per_glyph
    assert "$percent_per_glyph > 1.0" "width ($width) too great"
    debug "percent_per_glyph=$percent_per_glyph"
    
    debug "glyph_count=$glyph_count"
    myBc "scale=0;$width*$percent/100;scale=20" value
    
    printf -v value %.0f $value

    printf -v rhs %*s $((width-value))
    printf -v rhs %s "${rhs// /${theme[off]}}"
    printf -v out %*s $value
    if $reverse ;then
	printf -v out "%s%s" "$rhs" "${out// /${theme[on]}}"
    else
	printf -v out "%s%s" "${out// /${theme[on]}}" "$rhs"
    fi    
    echo $newline "$out"
}
show_theme_entry() {
    local -a theme="(${themes[$1]})"
    local min="$2" max="$3" current="$4"
    
    debug "theme=$theme"
    debug "min=$min"
    debug "max=$max"
    debug "current=$current"
    
    myBc "$max - $min" range
    
    myBc "((${current} - ${min})/${range})" quotient
    myBc "$quotient * 100" percent
    
    glyph_count=${#theme[@]}
    
    debug "range=$range"
    debug "quotient=$quotient"
    debug "percent=$percent"
    debug "glyph_count=$glyph_count"

    # just ASCII please
    if $a11y ;then
	[ -z "$decimal_places" ] && decimal_places=$default_decimal_places
	printf -v val "%.*f" $decimal_places $percent
	echo $newline "$val"
	return
    fi
    
    if [ $glyph_count -eq 2 ] ;then
		#show_rating_theme "${theme[*]}" "$min" "$max" "$current" "$percent"
	show_rating_theme $1 "$min" "$max" "$current" "$percent"
	return
    fi

    myBc "100/$glyph_count" percent_per_glyph
    debug "percent_per_glyph=$percent_per_glyph"
    
    assert "$percent_per_glyph > 1.0" "width ($width) too great"

    # with this scheme, assuming current value is 0-100 and theme has 10
    # elements:
    #
    # current   glyph from theme
    #
    # 0-19      1st
    # 20-29     2nd
    # 30-39     3rd
    #  :
    # 90-99     9th
    # 100       10th
    myBc "$quotient*$glyph_count-1" i
    [ -z "${i%-*}" ] && i=0
    printf -v i "%.0f" ${i%.*}
    ((i>glyph_count-1))&&i=$((glyph_count-1))
    
    debug "index=$i"
    content=${theme[i]}
    bc_test "$current == 0" && $zero_as_space && content=' '
    echo $newline "$content"
}

check_a11y

while getopts "abc:de:hilm:npqrt:u:w:x:" opt  ;do
  case "$opt" in
    a) a11y=true ;;
    b) zero_as_space=true ;;
    c) current=${OPTARG//−/-} ;;
    d) debug_enabled=true ;;
    e) decimal_places=$OPTARG ;;
    h) usage;exit 0 ;;
    i) invert=true ;;
    l) list=true ;;
    m) min=${OPTARG//−/-} ;;
    n) newline=-n ;;
    p) permissive=true ;;
    q) quiet=true ;;
    r) reverse=true ;;
    t) theme=$OPTARG ;;
    u) user_theme="$OPTARG" ;;
    w) width=$OPTARG ;;
    x) max=${OPTARG//−/-} ;;
  esac
done
shift $[$OPTIND-1]

runBackBc

if [ "$user_theme" ] ;then
    user_theme="${user_theme//\\/\\\\}"
    user_theme="${user_theme//\#/\\\#}"
    user_theme="${user_theme//\[/\\[}"
    user_theme="${user_theme//\]/\\]}"
    user_theme="${user_theme//\(/\\\(}"
    user_theme="${user_theme//\)/\\\)}"
    user_theme="${user_theme//&/\\&}"
    user_theme="${user_theme//;/\\;}"
    user_theme="${user_theme//</\\<}"
    user_theme="${user_theme//>/\\>}"
    user_theme="${user_theme//|/\\|}"
    user_theme="${user_theme//\~/\\~}"
    user_theme="${user_theme//\*/\\*}"
    user_theme="${user_theme//\$/\\$}"
    user_theme="${user_theme//\"/\\\"}"
    user_theme="${user_theme//\'/\\\'}"
    user_theme="${user_theme//\`/\\\`}"
    elements=($user_theme)
    elements=${#elements[@]}

    assert "$elements >= 2" "user themes need >= 2 values"
    
    # create new theme
    name="user_${elements}"
    themes[$name]="$user_theme"
    
    [ -z "$theme" ] && theme=$name
fi

if $list ;then
    if [ -z "$theme" ] ;then
	list_themes && exit 0
    else
	theme_valid "$theme" || die "invalid theme: $theme"
	$quiet || echo "Listing theme '$theme'"
	list_theme $theme && exit 0
    fi
fi

[ -z "$min" ]   && min=$min_default
[ -z "$max" ]   && max=$max_default
[ -z "$width" ] && width=$width_default

assert "$min <= $max" "minimum ($min) > maximum ($max)"
assert "$min != $max" "minimum ($min) == maximum"

[ -z "$current" -a ! -z "$1" ] && current="$1"

if [ -z "$current" ] ;then
    error "must specify current value"
    usage
    exit 1
fi

if $permissive ;then
    bc_test "$current < $min" && current=$min
    bc_test "$current > $max" && current=$max
else
    assert "$current >= $min" "current ($current) < minimum ($min)"
    assert "$current <= $max" "current ($current) > maximum ($max)"
fi

[ -z "$theme" ] && theme=$theme_default

theme_valid "$theme" || die "invalid theme: $theme"

show_theme_entry "$theme" $min $max $current

exit 0
